#!/usr/bin/python3
import os
import sys
import numpy as np
import copy
from keras.applications.resnet50 import ResNet50
from keras.applications.resnet50 import preprocess_input, decode_predictions
from keras.applications import VGG16
from keras.layers import Dense, Dropout, Activation, Flatten
from keras.models import Model, Sequential
from keras.optimizers import SGD 
from keras.preprocessing import image
from keras.preprocessing.image import ImageDataGenerator
from keras.callbacks import Callback, EarlyStopping
from keras.models import model_from_json
from keras.utils import np_utils
from sklearn.model_selection import train_test_split
from sklearn.model_selection import KFold
import pdb
import csv
import pathlib
import my_settings

opt = SGD(lr=1e-4, momentum=0.9)
X_paths = []
data_orig = {}
X = []
Y = []
output_manifest = []

def save_model( model, tag=None ):
    """
    The save_model function saves a keras neural network to disk as a pair of files: a json file that represents the structure and and h5 file that represents the wieghts.
    The function takes 2 parameters: model and tag. 
    - model is the keras model to save.
    - tag gets included in the filenames to provide some 'uniqueness' and helpful naming.
    The files are saved into the my_settings.trained_models_path directory. If the directory doesn't exist, it gets created.
    If the files already exist, they are replaced without notification. This could be undesirable behavior.
    """
    model_json = model.to_json()
    name = "model"
    if tag:
        name = name + "_" + tag
    pathlib.Path( my_settings.trained_models_path ).mkdir( exist_ok=True )
    model_file = os.path.join( my_settings.trained_models_path, name+".json" )
    with open( model_file, "w" ) as json_file:
        json_file.write(model_json)
    # serialize weights to HDF5
    weights_file = os.path.join( my_settings.trained_models_path, name + ".h5" )
    model.save_weights( weights_file )
    global output_manifest
    output_manifest.append( model_file )
    output_manifest.append( weights_file )
    print("Saved model to disk")
    
def get_flow_generator( test=False ):
    """
    The get_flow_generator function creates and returns an ImageDataGenerator object, which is the Keras mechanism for accessing large numbers of labelled data for training (or testing).
    The only input param is 'test' which is a boolean. The idea here is that if you're training and testing on the same dataset, you can access the train or test partition of the data.
    The path to the data is assumed to be in my_settings.data. It's not a param to the function.
    The ImageDataGenerator has many powerful arguments. They are currently all hardcoded. The imagine mean is also currently hardcoded.
    The function returns the generator which can then be passed to train (fit) or test code.
    """
    trainPath = my_settings.data
    trainAug = ImageDataGenerator( validation_split=my_settings.validation_split )
    #trainAug = ImageDataGenerator()
    mean = np.array([105.0, 110.0, 105.0], dtype="float32")
    trainAug.mean = mean
    print( "Going to 'flow-from-directory " + trainPath )
    data_subset = None # this is complicated for reasons that need to be explained
    if my_settings.mode == "test":
        data_subset='validation' 
    elif my_settings.mode == 'train':
        data_subset='training' 
    generator = trainAug.flow_from_directory(
        trainPath,
        class_mode="categorical",
        target_size=(224, 224),
        color_mode="rgb",
        shuffle=(not test), # This seems broken. TBD.
        batch_size=(32 if my_settings.mode=="train" else 1),
        follow_links=True,
        subset=data_subset
    )
    return generator

def get_new_vgg_model():
    # ResNet50 seems to be really well suited for prediction and easy decoding.
    # VGG16 seems well suited for feature vector work. 
    # include_top ( final/output classification layer ) is good for prediction.
    # false is good if you want to add your own classification schema.
    """
    The get_new_vgg_model function gets the trained VGG16 model which can be used as a base model for tuning.
    """
    base_model = VGG16( weights='imagenet', include_top=False, input_shape=(224, 224, 3))
    #model = ResNet50( weights='imagenet', include_top=True input_shape=(224, 224, 3))
    return base_model 
   
def extend_model_with_classification_layer( base_model, num_classes, freeze=False ):
    # Now we're going to add our own fully connected 3-value classificaiton softmax 
    # final (head) layer, and add it to the base model. Build it piece by piece.
    # NOTE: I'm surprised we replace the penultimate layer as well. Surely it had values
    # from imagenet training?
    # Lock down (freeze) vgg layers
    """
    The extend_model_with_classification_layer function takes a base_model (input parameter) and extends it with a classification layer based on num_classes which is also an input parameter.
    An optional input param is freeze which if True, locks down the base_model weights such that only the new weights will be subject to training.
    """
    if freeze:
        print( "We are freezing the weights in the trained imagenet model layers." )
        for layer in base_model.layers:
            layer.trainable = False
    newModel = base_model.output 
    newModel = Flatten(name="flatten")(newModel)
    newModel = Dense(4096, activation="relu")(newModel)
    newModel = Dropout(0.5)(newModel)
    newModel = Dense(num_classes, activation="softmax")(newModel)
    
    # place the head FC model on top of the base model (this will become
    # the actual model we will train)
    model = Model(inputs=base_model.input, outputs=newModel)
    #optimizers.RMSprop(lr=2e-4)
    model.compile(loss="categorical_crossentropy", optimizer=opt, metrics=["accuracy"])
    model.summary()
    return model

def extend_model_with_regression_layer( base_model, freeze=False ):
    # Now we're going to add our own fully connected regression 
    # final (head) layer, and add it to the base model. 
    # from imagenet training?
    # Lock down (freeze) vgg layers
    """
    The extend_model_with_regression_layer function takes a base_model (input parameter) and extends it with a regression layer.
    An optional input param is freeze which if True, locks down the base_model weights such that only the new weights will be subject to training.
    It prints a (text) summary/visualization of the network to the console. 
    Ituses mean_squared_error as the loss function which was discovered by trial-and-error to be very important.
    """
    #base_model.summary()
    newModel = None
    if freeze:
        for layer in base_model.layers:
        #for layer in base_model.layers[:-1]:
            layer.trainable = False
    if "activation" in base_model.layers[-1].get_config() and base_model.layers[-1].get_config()["activation"] == "softmax":
        print( "Loaded model has a classification layer that needs to be replaced." )
        # Strip the classification layer and the 4096 layer
        #base_model.summary()
        #base_model = base_model.layers.pop() # pop another layer
        newModel = Sequential()
        for layer in base_model.layers[:-1]:
            newModel.add( layer )
        #newModel = base_model.output
        #newModel = Dense( 1, kernel_initializer='uniform', activation="linear", name="dense_out")(newModel) # fumbling around here trying to replace softmax layer with second dense layer.
        newModel.add( Dense( 1, kernel_initializer='uniform', activation="linear", name="dense_out") )
    else: 
        print( "Loaded model has no classification layer that needs to be replaced." )
        newModel = base_model.output
        newModel = Flatten(name="flatten")(newModel)
        newModel = Dense(4096, kernel_initializer='uniform', activation="relu")(newModel)
        newModel = Dropout(0.5)(newModel)
        newModel = Dense( 1, kernel_initializer='uniform', activation="linear")(newModel)
 
        # place the head FC model on top of the base model (this will become
        # the actual model we will train)
        newModel = Model(inputs=base_model.input, outputs=newModel)

    #optimizers.RMSprop(lr=2e-4)
    #model.compile( loss="binary_crossentropy", optimizer='adam', metrics=['accuracy'] )
    newModel.compile( loss='mean_squared_error', optimizer='adam', metrics=['accuracy'] )
    newModel.summary()
    return newModel

def train_and_save( tag=None, num_classes = 3, freeze=False ):
    """
    The 'train_and_save' function starts with an existing VGG16 model (using the get_new_vgg_model function), 
    extends with a classification layer (using the extend_model_with_classification_layer), 
    gets a generator (with the get_flow_generator fuction), and trains the model from the data.
    There is a stopAtLossValue callback which stops training once a metric is achieved (epoch-to-epoch loss delta).
    The trained model is saved to disk using the save_model function.
    """
    model = get_new_vgg_model()
    model = extend_model_with_classification_layer( model, num_classes, freeze )
    trainGen = get_flow_generator()
    
    # Copying a stop-at-loss callback solution from s/o. Did not get this to work but I still find it interesting.
    class stopAtLossValue(Callback):
        def on_batch_end(self, batch, logs={}):
            THR = 0.03 #Assign THR with the value at which you want to stop training.
            if logs.get('loss') <= THR:
                self.model.stop_training = True
    
    STEP_SIZE_TRAIN=trainGen.n//trainGen.batch_size
    print( "STEP_SIZE_TRAIN calculated as " + str( STEP_SIZE_TRAIN ) )
    #STEP_SIZE_VALID=valid_generator.n//valid_generator.batch_size
    mycallbacks = [
    	#stopAtLossValue
    	EarlyStopping(monitor='loss', min_delta=0.01, verbose=1)
    ]
    model.fit_generator(generator=trainGen,
                        steps_per_epoch=100,
                        #validation_data=valid_generator,
                        #validation_steps=STEP_SIZE_VALID,
                        epochs=100,
                        callbacks=mycallbacks)
    
    save_model( model, tag )
    
def train_and_save_binary_classifier( tag=None, freeze=False ):
    """
    The 'train_and_save_binary_classifier' is a version of train_and_save with the number of classes hardcoded to 2.
    """
    return train_and_save( tag, num_classes=2, freeze=freeze )

def load_trained_model( tag=None ):
    # load json and create model
    """
    The 'load_trained_model' function loads an existing model basad on input parameter tag and trained_model_path from my_settings
    and returns it to the caller.
    """
    name = "model"
    #if tag:
    if my_settings.model is not None:
        name = name + "_" + my_settings.model # something's broken here. tag is required but not used?!?
    print( "load_trained_model: name = {}.".format( name ) )
    json_file = open( os.path.join( my_settings.trained_models_path, name+".json" ), 'r' )
    loaded_model_json = json_file.read()
    json_file.close()
    loaded_model = model_from_json(loaded_model_json)
    # load weights into new model
    loaded_model.load_weights( os.path.join( my_settings.trained_models_path, name+".h5" ) )
    loaded_model.compile(loss='categorical_crossentropy',
                         optimizer=opt,
                         metrics=['accuracy']) 
    print("Loaded model from disk")
    return loaded_model
    
def flow_feature( tag ):
    """
    The 'flowtest' function is the longest functin in this module. It loads an existing trained model from disk and
    does predictions on each file in the specified test dataset. It calculates an average accuracy based on whether the
    predicted label matches the actual. ALSO, it calculates and saves feature vectors based on removing the final 
    classificaiton layer of the main neural net/model. The feature vectors are saved in a separate file from the 
    feature filenames used. These can be stiched together in post via unix command paste.
    """
    if tag is None and my_settings.model:
        tag = my_settings.model
    print( "Loading pre-trained model based on tag: " + tag )
    feature_model = copy.deepcopy( load_trained_model( tag ) )
    feature_model.layers.pop() # not interested in classification
    feature_model = Model(inputs=feature_model.inputs, outputs=feature_model.layers[-1].output)
    print( "Creating feature model. Getting flow generator [can be very slow for large datasets]." )

    testGen = get_flow_generator()
    testGen.reset()
    tests = 0
    
    features_arr = None
    num_images = len( testGen.filenames )
    print( "Writing feature vectors based on tag: {}.".format( tag ) )
    features_filename = "features"
    if tag:
        features_filename += ( "_" + tag )

    print( "Starting to loop through all {} images to test and calc feature vectors.".format( num_images ) )
    for inputs_batch, labels_batch in testGen:
        if tests >= num_images:
            break

        features = feature_model.predict(inputs_batch)
        arr = np.array( features )
        with open( features_filename + ".csv", "a" ) as csvout:
            np.savetxt( csvout, arr, '%s', ',' ) 
            #print( testGen.filenames[tests], str( features[i] ) ) 
        tests += 1

    # worry about corresponding input filenames later 
    with open( features_filename + "_filenames.txt", "w+" ) as input_files:
        for filename in testGen.filenames:
            input_files.write( pathlib.PurePath(filename).name + "\n" )

    global output_manifest
    output_manifest.append( features_filename + "_filenames.txt" )
    output_manifest.append( features_filename + ".csv" )
    
def flowtest( tag ):
    # This will go in a function so it's a separate runtime option. For now just
    # getting code that works.
    # We want 1 model for test statistics and 1 for feature extraction, but probably not at the same time!
    """
    The 'flowtest' function is the longest functin in this module. It loads an existing trained model from disk and
    does predictions on each file in the specified test dataset. It calculates an average accuracy based on whether the
    predicted label matches the actual.
    """
    if tag is None and my_settings.model:
        tag = my_settings.model
    print( "Loading pre-trained model based on tag: " + tag )
    test_model = load_trained_model( tag ) 
    testGen = get_flow_generator()
    testGen.reset()
    tests = 0
    corrects = 0 
    num_images = len( testGen.filenames )

    print( "Starting to loop through all {} images to test.".format( num_images ) )
    for inputs_batch, labels_batch in testGen:
        if tests >= num_images:
            break

        preds = test_model.predict(inputs_batch) 
        for i in range( len( preds ) ):
            tests += 1
            pred = np.argmax(preds[i])
            label = np.argmax(labels_batch[i])
            if pred == label:
                corrects += 1
    print( "Test Accuracy {}: {}.".format( tests, 1.0*corrects/tests ) )

def do_regression_train_and_score( regression_model, X_train, Y_train, X_test, Y_test ):
    score_untrained = regression_model.evaluate( X_test, Y_test )
    print( "Score (Before Tuning): {}.".format( 100*dict(zip( regression_model.metrics_names, score_untrained ))["acc"]) )
    regression_model.fit( X_train, Y_train, nb_epoch=15, batch_size=10, verbose=1 )
    #model.fit( X, Y, nb_epoch=200, batch_size=10 )
    score_trained = regression_model.evaluate( X_test, Y_test )
    print( "Score (After): {}.".format( 100*dict(zip( regression_model.metrics_names, score_trained ))["acc"]) )
    #return score_untrained, score_trained
    one_by_one = True
    errs = []
    if one_by_one:
        #for input_datum in X_paths:
        #for input_datum in X_test: 
        for idx in range(len(X_test)): 
            #my_settings.data = input_datum
            input_datum = X_test[idx].reshape(-1, 224, 224, 3)
            #pdb.set_trace()
            predict = predict_single( regression_model, input_datum )
            actual = Y_test[idx]
            err = abs(predict-actual)
            errs.append( err )
            print( "Predicted = {}, Actual = {}, err = {}, mean_abs_err = {}.".format( predict, actual, err, sum(errs)/len(errs) ) )
    return dict(zip( regression_model.metrics_names, score_untrained ))["acc"], dict(zip( regression_model.metrics_names, score_trained ))["acc"]

def train_regression_network( model_fn, tag=None ):
    """
    The 'train_regression_network' function is an alternative to classification based training. It doesn't use 'flow-from-directory'
    at all, because that is based on classification labels. It uses a scalar value presumed to be in the final column of the input csv.
    It takes as a parametera model that better have a top/final layer that is NOT a classification layer. It also takes 'tag' as an 
    optional input parameter which is used for loading a named model. It also saves the trained model as per usual.
    An additional piece of functionality is that it calculates the performance of the network on the dataset before traiing and after
    so the user can see if the prediction accuracy after training is an improvement over the untrained net.
    """
    orig_model = model_fn()
    #X = np.empty([0,400,400,3])
    #with open( sys.argv[-1], "r" ) as csvfile:
    with open( my_settings.data, "r" ) as csvfile:
        reader = csv.reader( csvfile )
        next( reader, None ) # skip first row (header)
        for row in reader:
            try:
                X_paths.append( row[0] )
                #print( "Saving image path: {}.".format( row[0] ) )
                #Y.append( float(row[-1]) )
                dhs_val = 1000*float(row[-1])
                #dhs_val = round(dhs_val,0)
                #Y.append( 0 if dhs_val < 0.5 else 1 )
                threshold_labels = False
                if threshold_labels:
                    dhs_val = 0 if dhs_val < 0.5 else 1
                Y.append( dhs_val )
                data_orig[row[0]] = dhs_val
            except Exception as ex:
                print( "Exception ingesting and parsing input data.\n" + str(ex) )
    # load images
    for image_path in X_paths:
        #print( "Attempting to load {}.".format( image_path ) )
        sys.stdout.write( "." )
        img = image.load_img( image_path, target_size=(224,224) )
        img = image.img_to_array(img)
        img = np.expand_dims(img, axis=0)
        img = preprocess_input(img)
        X.append( img )
    print( "Training...")
    X_batch = np.vstack(X)
    Y_batch = np.vstack(Y)
    k_fold = True
    new_model = model_fn()
    if k_fold:
        #new_model = copy.deepcopy( orig_model )
        #new_model = orig_model
        scores_untrained = []
        scores_trained = []
        kf = KFold(n_splits=4)
        for train_idx, test_idx in kf.split( X_batch ):
            X_train, X_test = X_batch[train_idx], X_batch[test_idx]
            Y_train, Y_test = Y_batch[train_idx], Y_batch[test_idx]

            #X = X.reshape(X.shape[0], 3, 224, 224) nope, network expects it in 224,224,3
            #Y = np_utils.to_categorical( Y, 2 )
            score_untrained, score_trained = do_regression_train_and_score( new_model, X_train, Y_train, X_test, Y_test )
            scores_untrained.append( score_untrained )
            scores_trained.append( score_trained )
        #pdb.set_trace()
        print( "Mean score (before) = {}.".format( sum(scores_untrained)/len(scores_untrained)) )
        print( "Mean score (after) = {}.".format( sum(scores_trained)/len(scores_trained)) )
    else:
        X_train, X_test, Y_train, Y_test = train_test_split( X, Y, test_size=0.2 )
        do_regression_train_and_score( new_model, X_train, Y_train, X_test, Y_test )

    if my_settings.save_trained_model:
        save_model( new_model, tag )

def predict_single( model_for_predict=None, img=None ):
    """
    The 'predict_single' function loads an existing model, runs a single image through it, and displays the prediction.
    It is up to the user to know what the classes/labels are. The console output is just an array of probabiliies.
    """
    if model_for_predict is None:
        tag = my_settings.model
        print( "Loading pre-trained model based on tag: " + tag )
        model_for_predict = load_trained_model( tag )
    if img is None:
        img = image.load_img( my_settings.data, target_size=(224,224) )
        img = image.img_to_array(img)
        img = np.expand_dims(img, axis=0)
        img = preprocess_input(img)
    preds = model_for_predict.predict( img )
    #print( preds )
    return preds[0]
    

def get_tag(): 
    """
    The 'get_tag' function returns a tag string based on the data parameter specified by the user.
    """
    tag = my_settings.data if len(sys.argv) >= 2 else None # argv usage here should be retired
    tag = tag.strip( os.path.sep )
    if len(tag.split( os.path.sep ) ) > 1:
        tag = tag.split( os.path.sep )[-1]
    return tag

def get_vgg_model_and_extend_with_regression_top():
    model = extend_model_with_regression_layer( get_new_vgg_model() )
    return model

def load_trained_model_and_extend_with_regression_top():
    model = extend_model_with_regression_layer( load_trained_model() )
    return model

if __name__ == "__main__":
    """
    Can be invoked either by regular name 'train_imagenet_with_nl.py' for classification or as train_regression.py for regression.
    Classification:
    - train or test (or train then test)
        - train:
            - Loads VGG16 imagenet as starting point
                 - Freeze sets imagenet layers as non-trainable, only new fully connected layers are.
            - Saves trained model files, uses 'tag' as label and trained_model_path as directory
            - Can specify train/test split, defaults to 0.2
            - data specifies data to train on
        - test
            - Load model using 'tag' as label and trained_model_path as directory
            - data specifies data to test on; can test on separate dataset entirely or test split of training dataset
            - Generates feature vectors which are saved; tag used to generate name. Directory is (TBD)
            - single option classifies a single image instead of a metric across a dataset
        - NOTE: when training and testing on a single dataset, we want to shuffle the data but of course this needs to be done before
        splitting into train & test. I'm just worried that when we are 'flowtesting' for feature vector generation, we need to process
        the dataset in order.
    Regression:
        - train & test with kfold is default
        - Starts with VGG16
            - Specifying model starts with our own trained model
         - Freeze sets imagenet layers as non-trainable, only new fully connected layers are.
        - kfold is turn-off-able
    """
    print( sys.argv[0] )
    # let's try just storing all command-line args as key-value pairs in my_settings
    #my_settings.save_trained_model = True
    for arg in sys.argv[1:]:
        kv = arg.split( '=' )
        key = kv[0].strip("-")
        val = kv[1] if len(kv)==2 else 1
        print( "Found key {} and value {}.".format( key, val ) )
        my_settings.__dict__[key] = val

    if sys.argv[0].endswith( "train_regression.py" ):
        # New feature: If user invokes train_regression with model=<something>, assume they want to use
        # a pre-existing trained and saved model and load that up instead of the vgg model.
        model_fn = None
        if "model" in my_settings.__dict__:
            #model = load_trained_model() # or something
            model_fn = load_trained_model_and_extend_with_regression_top
        else:
            #model = get_new_vgg_model()
            model_fn = get_vgg_model_and_extend_with_regression_top
        #model = extend_model_with_regression_layer( model )
        #train_regression_network( model, get_tag() )
        train_regression_network( model_fn, get_tag() )
        sys.exit() # to avoid else and strange indentation

    if len(sys.argv) == 0: # the zero arguments mode was originally the train & test mode. data would have to be provided in my_settings.py
        #freeze = True if sys.argv[2] == "--freeze" else False
        train_and_save( get_tag(), 3, my_settings.freeze )
        flowtest() 
    elif sys.argv[1] == "--train":
        my_settings.mode = "train"
        tag = get_tag()
        #print( "Calling train_and_save_binary_classifier with tag={} and freeze={}.".format( tag, my_settings.freeze ) )
        #train_and_save_binary_classifier( tag, my_settings.freeze )
    elif sys.argv[1] == "--test":
        my_settings.mode = "test"
    elif sys.argv[1] == "--feature":
        my_settings.mode = "feature"
    else:
        print( "Usage: <script> [--train|--test|--feature]\nNo arguments means do train and test." )

    if my_settings.mode == "train":
        train_and_save( tag, 2 )
    elif my_settings.mode == "test":
        if "single" in my_settings.__dict__:
            print( predict_single() )
        else:
            print( "You chose to test a trained model on a dataset." )
            flowtest( None )
    elif my_settings.mode == "feature":
        flow_feature( None )

    with open( "output_manifest.txt", "w" ) as om_file:
        for om_filename in output_manifest: 
            om_file.write( om_filename + "\n" );

    if my_settings.mode == "feature":
        os.system( "merge_feature_outputs.sh" )
